"""
Created the 2022/09/28
v0.1 First version
@author: Nicolas Hardy

This file is part of Fermy.

    Fermy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Fermy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Fermy.  If not, see <https://www.gnu.org/licenses/>.
"""
__version__ = 0.1

import minimalmodbus
import datetime
from typing import List, Dict, Iterable, Tuple, Union
from . import errorslabtools


class M80:
    """Class that connect and drive a Mettler Toledo sensor mount
    transmitter connected to a PC with an USB/RS485 convertor
    Material.
    M80-SM transmietter 30530566
    5-pin data cable 2m 52300379
    USB/Rs485 e.g. SB485 Papouch Term485 On and SW4 On
    24V power supply e.g. MeanWell DR-15-24
    
    Wiring
    Open calbe end = function = connected to
    Brown = 24 VDC+ = positive output pin of the power supply
    black = 24 VDC- = negative output pin of the power supply
    blue = RS485- (B) = RxTx- pin of USB converter
    white = Rs485+ (A) = RxTx+ pin of USB converter
    gray = groud = groud pin of USB converter
    yellow = shield = connection to earth (power supply)
    
    from COM port information it is possible to find it with "pyserial" tool #python -m serial.tools.list_ports
    """
    def __init__(self, portcom, deviceaddress, baudrate: int=38400, bytesize: int=8,
                parity:str=minimalmodbus.serial.PARITY_NONE,
                stopbits:int=1, timeout: float=0.1): #Our construction method
        """Initialisation from port COM"""
        self.portcom = portcom
        self.baudrate = baudrate
        self.bytesize = bytesize
        self.parity = parity
        self.stopbits = stopbits
        self.timeout = timeout
        self.deviceaddress = deviceaddress
        self.connection = False
        self.instrument = None
        self.time = datetime.datetime.now()
        self.M1 = None
        self.M2 = None
        self.M3 = None
        self.M4 = None

    def stopcom(self):
        """Stop connection
        """
        try:
            self.instrument.serial.close() # stop the communication
        except Exception as err:
            raise errorslabtools.CommunicationError("Not connected\nPlease connect an instrument first with .startcom()")
        else:
            self.instrument.serial.close() # stop the communication
            self.connection = False
            print("Disconnection")
            
    def startcom(self):
        """Srat connection with the balance
        """
        if self.connection == False:
            try:
                self.instrument = minimalmodbus.Instrument(self.portcom, self.deviceaddress, mode="rtu")  # port name, slave address (in decimal) 49 for TQS4
            except IOError: # or Exception as err: ?
                raise errorslabtools.CommunicationError(f"No device answer probably due to a wrong port name ({self.portcom}).")
            else:
                self.instrument.serial.baudrate = self.baudrate
                self.instrument.serial.bytesize = self.bytesize
                self.instrument.serial.parity   = self.parity
                self.instrument.serial.stopbits = self.stopbits
                self.instrument.serial.timeout  = self.timeout        # in seconds
                self.instrument.clear_buffers_before_each_transaction = True
                self.connection = True
            self.testaddress()
        else:
            print("Already connected")

    def __str__(self) -> str:
        """Define printing method print(object)"""
        return f"M80 connection = {self.connection}\nM80 device address {self.deviceaddress} \
        \nlast read: {self.time}\nLast values {self.M1, self.M2, self.M3, self.M4}"
    
    def testaddress(self) -> bool:
        """Method to test if a device is connected
        at the requested address
        """
        if type(self.instrument) == minimalmodbus.Instrument:
            try:
                comok = self.instrument.read_register(0)
            except IOError: # or Exception as err: ?
                raise errorslabtools.CommunicationError(f"No device answer probably due to a wrong slave address {self.deviceaddress}")
            else:
                comok = self.instrument.read_register(0)
                if comok == 0:
                    print(f"A device is avaible with address {self.deviceaddress}")
                    return True
                else:
                    print("Please connect an instrument first with .startcom()")
    
    def read(self)-> Dict:
        """read all channels of the connected probe with reading datetime
        and return a Dict with  {"time","M1", "M2", "M3", "M4"}
        """
        if self.connection:
            self.time = datetime.datetime.now()
            try:
                self.M1 = self.instrument.read_float(100) # read mesurment M1 for me pH (Cf. doc p28)
                self.M2 = self.instrument.read_float(102) # read mesurment M2 for me °C
                self.M3 = self.instrument.read_float(104) # read mesurment M3 for me Volts (pH)
                self.M4 = self.instrument.read_float(106) # read mesurment M4 for me ORP (V)
            except IOError:
                raise errorslabtools.CommunicationError("No able to read the prob please check connection.")
            else:
                print(f"time: {self.time}\nM1: {self.M1}\nM2: {self.M2}\nM3: {self.M3}\nM4: {self.M4}\n")
                return {"time": self.time,
                        "M1" : self.M1,
                        "M2" : self.M2,
                        "M3" : self.M3,
                        "M4" : self.M4,
                        }
        else:
            print('Please first used .startcom()') 
            return {}



"""
Created the 2022/10/04
v0.1 First version
v0.2 Bugfix with Epoch2 and fakedata name
v0.3 add Tecan reader 96 and 384
v0.4 add Logphase600 96 BioTek & add exclusion of rows with nan
@author: Nicolas Hardy



This file is part of Fermy.

    Fermy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Fermy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Fermy.  If not, see <https://www.gnu.org/licenses/>.
"""

__version__ = 0.4

import pandas as pd
from typing import List, Dict, Iterable, Tuple, Union
import math
from . import errorslabtools

class MTPReader():
    """Class that allow to read MTP data need filepath and readername
    """
    
    def __init__(self, filepath:str, readername:str):
        """Initialisation of reader
        """
        self.readername = readername
        self.plate = pd.DataFrame()
        self.filepath = filepath
        self.dicoreader = {"Epoch2" : self.readepochtwo,
                            "fakeMTP" : self.readfakemtp,
                            "Tecan96" : self.readtecan96,
                            "Tecan384" : self.readtecan384,
                            "Logphase" : self.readlogphasesixhundr
                            } # linker between MTP and reading function
    
    def convstringtohours(self, string: str) -> float:
        """"Function to transform a string like hours:minutes:seconds
        or aaa-mm-dd hh:mm:ss to a float of hours
        return float
        """
        if len(string.split(" ")) == 1:
            spliting = string.split(':')
            timefloat = float(spliting[0])+float(spliting[1])/60+float(spliting[2])/3600
        else:
            spliting = string.split(" ")
            timefloat = float(spliting[0].split('-')[2])*24
            timefloat = timefloat + float(spliting[1].split(':')[0])
            timefloat = timefloat + float(spliting[1].split(':')[1])/60
            timefloat = timefloat + float(spliting[1].split(':')[2])/3600
        return timefloat
    
    def readepochtwo(self):
        """Function to return formated dataframe from reader Epoch
        index is time float in hours
        """
        
        rawtable = pd.read_excel(self.filepath, sheet_name=0)
        numline = int(rawtable[rawtable["Unnamed: 0"]=="Layout"].index.values) +2
        del rawtable
        tabledata = pd.read_excel(self.filepath, sheet_name=0, usecols="B:CU", index_col=None, skiprows=numline+13, skipfooter=36, dtype={"Time":str})  #data robot
        tabledata["Time"] = tabledata["Time"].apply(self.convstringtohours)  # converrt time from hh:mm:ss to float
        tabledata.index = tabledata["Time"] - tabledata["Time"].iloc[0] # set index
        tabledata.drop(["Time", "T° 600"], axis=1, inplace=True, errors = "ignore")  # Drop Time columns
        tabledata = tabledata.applymap(lambda data: data.strip("*") if isinstance(data, str) else data)  # clean *
        tabledata = tabledata.astype('float64')
        tabledata.dropna(axis=0, how="all", subset=tabledata.columns[2:], inplace=True)  # clear empty data
        tabledata.index.name = "Time"
        return tabledata
    
    def readtecan96(self):
        """Function to return formated dataframe from reader Tecan with 96
        index is time float in hours
        """
        rawtable = pd.read_excel(self.filepath, sheet_name=0)
        numlinedata = int(rawtable[rawtable.iloc[:,0] == "Cycle Nr."].index.values)
        numlineenddata = int(rawtable[rawtable.iloc[:,0] == "End Time"].index.values[-1])
        numlineendtoskip = rawtable.index.max() - numlineenddata + 2
        del rawtable
        tabledata = pd.read_excel(self.filepath, sheet_name=0, usecols="B:CU", index_col=None, skiprows=numlinedata+1, skipfooter=numlineendtoskip)  #data plate 96
        tabledata["Time"] = tabledata["Time [s]"]/3600 # conversion to hours
        tabledata.index = tabledata["Time"]  # set time as index
        tabledata.drop(["Time", "Temp. [°C]", "Time [s]"], axis=1, inplace=True, errors = "ignore")  # Drop Time columns
        return tabledata
    
    def readtecan384(self):
        """Function to return formated dataframe from reader Tecan with 96
        index is time float in hours
        """
        rawtable = pd.read_excel(self.filepath, sheet_name=0)
        numlinedata = int(rawtable[rawtable.iloc[:,0] == "Cycle Nr."].index.values)
        numlineenddata = int(rawtable[rawtable.iloc[:,0] == "End Time"].index.values[-1])
        numlineendtoskip = rawtable.index.max() - numlineenddata + 2
        del rawtable
        tabledata = pd.read_excel(self.filepath, sheet_name=0, usecols="B:NW", index_col=None, skiprows=numlinedata+1, skipfooter=numlineendtoskip)  #data plate 384
        tabledata["Time"] = tabledata["Time [s]"]/3600 # conversion to hours
        tabledata.index = tabledata["Time"]  # set time as index
        tabledata.drop(["Time", "Temp. [°C]", "Time [s]"], axis=1, inplace=True, errors = "ignore")  # Drop Time columns
        return tabledata
        
    def readlogphasesixhundr(self):
        """Function to return formated dataframe from Logphase600
        from BioTek Agilent
        index is time float in hours
        """
        with open(self.filepath,"r") as filetemp:
            lines = filetemp.readlines()
            indexfile = 0
        for line in lines:
            if "Time\t" in line:
                numline = indexfile
                # print(f"{indexfile} in ''{line}'")
            indexfile += 1
        tabledata = pd.read_csv(self.filepath, skiprows=numline, encoding='cp1252', sep="\t", skip_blank_lines=True, decimal=",")
        tabledata.index = tabledata["Time"].apply(self.convstringtohours)  # converrt time from hh:mm:ss to float
        tabledata.drop(["Time"], axis=1, inplace=True, errors = "ignore")  # Drop Time columns
        return tabledata
        
    def readfakemtp(self):
        """Function to return a fake formated dataframe
        """
        # time 5 minutes steps in hours for 6 hours
        time = [time/60 for time in range(0, 60*6, 5)]
        lagtime = time[20]  # 1.66 h
        fakedataset = [0.01]*20+[0.01*math.exp(0.5*(time-lagtime)) for time in time[20:]]
        datadico = {}
        for line in ["A", "B", "C","D", "E", "F", "G", "H"]:
            for column in range(1,13):
                datadico[f"{line}{column}"] = fakedataset
        tabledata = pd.DataFrame(data= datadico, index=time)
        tabledata.index.name = "Time"
        return tabledata
    
    def readMTP(self) -> pd.DataFrame:
        """Load a file and provide a formated DataFrame
        """
        #read the input file if the reader name is known 
        if self.readername in self.dicoreader.keys():
            try:
                self.plate = self.dicoreader[self.readername]()
            except Exception as err:
                raise errorslabtools.MTPError(f"Wrong device '{self.readername}' or corrupted file")
        self.plate.dropna(axis=0, how="any", inplace=True) # clean row with missing data (useless)
        return self.plate
    
    def readerslist(self) -> List[str]:
        """Provide the names of readers avaible
        """
        listofreaders = list(self.dicoreader.keys())
        return listofreaders




"""
Created the 2021/10/13
v0.0 First version
v0.1 Add link with pandasfermy
v0.2 Add control chart graph
v0.3 Add powerfoce option to control chart and bug fix if no vialtion of rules
v0.4 update of function description
        for ccplot ==> no show if save path set / add option to unclean data violation
v0.5 Add multiy chart method and chart are no shown if save path set
v0.6 Add groupby option multiyaxis chart
v0.7 bug fix for more than two bioreactors dataset multiyplot grouped
v0.8 add colorby option to ccplot
v0.9 add option to comput for ccplot LCL and UCL and mean from a pd.serie
v1.0 add option to set x axis title

Next step change comportement of pah/filepath and use pathlib instead of os.path

To implement for multiy chart:
    * option to add animation if feasable
    * option to plot some colmns with dot instead line (input list of columns)
    datatodots:List[str]=[]

@author: Nicolas Hardy

This file is part of Fermy.

    Fermy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Fermy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Fermy.  If not, see <https://www.gnu.org/licenses/>.
"""

__version__ = 0.9


import pandas as pd
from typing import List, Dict, Iterable, Tuple, Union
import plotly.express as px
import plotly.graph_objects as go  # only for multi y plot
import os
import argparse
import numpy as np

#layout
layoutdic = dict(modebar_add=["v1hovermode", "toggleSpikeLines"],
                    title='Fermentation data',
                    xaxis_title="Time (in hours)",
                    yaxis_title="Growth proxy",
                    font=dict(
                        family='Linux Libertine, Times New Roman',
                        size = 20,
                        color='#000'
                    ),
                    legend=dict(
                        title=dict(text="Bioreactors",side="top"),
                        x=0,
                        y=-0.2,
                        orientation="h",
                        font=dict(
                            family='Linux Libertine, Times New Roman',
                            size = 10,
                            color='#000'
                        ),
                        traceorder="normal"  # "normal" or "grouped"
                    ),
                        updatemenus=[
                            dict(
                                buttons=list([
                                    dict(label="Linear",  
                                        method="relayout", 
                                        args=[{"yaxis.type": "linear", "yaxis.title" : f"Growth proxy"}]),
                                    dict(label="Log", 
                                        method="relayout", 
                                        args=[{"yaxis.type": "log", "yaxis.title" : f"ln(Growth proxy)"}])
                                  ]),
                                x=0.5,
                                y=1.5,
                                  ),
                                  {'type': 'buttons',
                    "showactive":True,
                    "x" : 0.6,
                    "y" : 1.5,
                    'buttons': [{'label': 'Legend',
                                'method': 'relayout',
                                'args': ['showlegend', True],
                                'args2': ['showlegend', False]}],
                    }
                                  
                                  
                                  ]
                )



class ParseKwargs(argparse.Action):
    """Class to add deal with dict as arg in argparse
    """
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, dict())
        for value in values:
            key, value = value.split('=')
            getattr(namespace, self.dest)[key] = value



def basicplot(data: pd.DataFrame, path:str=""):
    """Plot all columns of a Dataframe as Y and index for x
    """
    fig = px.line(data, x=data.index, y=data.columns, markers=True)
    fig.update_layout(layoutdic, template="simple_white")
    if path != "":
        fig.write_html(os.path.splitext(path)[0]+"_sfgraph.html")
    else:
        fig.show()

def defxaxislen(list_of_yaxis: List, yspace: float):
    """Give the value for the max xaxis range from the list of y axis linked to the number
    """
    lenlistyaxis = len(list_of_yaxis)
    if lenlistyaxis > 2:
        xaxislen = (1-(lenlistyaxis-2)*yspace) #Range max for x depends on the yxais number exemple for 4 it is 0.9.
    else:
        xaxislen = 1
    return xaxislen


def multiyplot(data: pd.DataFrame, path:str="", yspace:float=0.05, sizep:int=16, groupby:Union[None,str]=None, xtitle:Union[None,str]=None):
    """Plot all columns of a Dataframe with multiY axis and index for x
    
    Input:
    -----
    DataFrame
    
    
    Parameters:
    ----------
    path: is provided the graph will be saved in this path
    
    yspace: allows users to custom space between yaxis
    
    sizep: allows users to custom font size
    
    groupby: used character to build group for plotting e.g. groupby="-" for pH-Biroeactor1 => group pH and subdataset bioreactor1
    
    xtitle: allows to custom x title
    
    Output:
    ------
    No output only graph
    """
    fig = go.Figure()  # creat an empty fig
    
    if groupby:
        #groupe construction dico
        headerslist = data.columns.tolist()
        headerslist.sort()  # sort list
        dicolistgroup ={}
        for header in headerslist:
            dicolistgroup[header] = header.split(groupby)[0]
        listofgroup = list(set(dicolistgroup.values()))
        listofgroup.sort()  # sort list
        rangemaxXaxis = defxaxislen(listofgroup, yspace) #variable size of x axis
        
        assert rangemaxXaxis>0, "too many columns please drop useless columns and/or reduce yspace paramter"
        
        fig.update_layout(xaxis=dict(domain = [0,rangemaxXaxis], #place en % des x
                        title=data.index.name,
                        title_standoff=0,
                        ),
                        title_text="Fermentation multiple y-axes",
                        template="simple_white",
                        legend=dict(x=0,
                                    y=-0.1,
                                    tracegroupgap = 0,
                                    orientation="h",
                                    font=dict(
                                        family='Times New Roman',
                                        size = sizep,
                                        color='#000')
                                    ),
                        font=dict(
                            family='Linux Libertine, Times New Roman',
                            size = sizep,
                            color='#000'
                        ),
                        modebar_add=["v1hovermode", "toggleSpikeLines"],
                        )
        indexgroup = 0
        legendgroup = None
        for column in data.columns.sort_values():
            if indexgroup == 0:
                if legendgroup == None:
                    legendgroup = dicolistgroup[column]
                    fig.add_trace(go.Scatter(x=[data.index[0]], y=[data[column][0]], name=legendgroup, marker=dict(color="Black", size=0.01), visible='legendonly', legendgroup=legendgroup, showlegend=True))  #Q&D no better solution today
                    fig.add_trace(go.Scatter(x=data.index, y=data[column], name=column, legendgroup=legendgroup, hovertemplate="x: %{x:.2f}<br>y: %{y:.2f}", showlegend=False))
                    fig.layout["yaxis"] = dict(title=legendgroup, title_standoff=0, title_font={"size" : sizep*rangemaxXaxis})
                    
                elif legendgroup == dicolistgroup[column]:
                    fig.add_trace(go.Scatter(x=data.index, y=data[column], name=column, legendgroup=legendgroup, hovertemplate="x: %{x:.2f}<br>y: %{y:.2f}", showlegend=False))  
                    fig.layout["yaxis"] = dict(title=legendgroup, title_standoff=0, title_font={"size" : sizep*rangemaxXaxis})
                else:
                    indexgroup+=1
                    legendgroup = dicolistgroup[column]
                    fig.add_trace(go.Scatter(x=[data.index[0]], y=[data[column][0]], name=legendgroup, marker=dict(color="Black", size=0.01), visible='legendonly', legendgroup=legendgroup, showlegend=True, yaxis=f"y{indexgroup+1}"))  #Q&D no better solution today
                    fig.add_trace(go.Scatter(x=data.index, y=data[column], name=column, yaxis=f"y{indexgroup+1}", legendgroup=legendgroup, hovertemplate="x: %{x:.2f}<br>y: %{y:.2f}", showlegend=False))
                    fig.layout[f"yaxis{indexgroup+1}"] = dict(title=legendgroup, anchor="x", overlaying="y", side="right", position=1, title_standoff=0, title_font= {"size" : sizep*rangemaxXaxis})
                    
            elif indexgroup == 1:
                if legendgroup == dicolistgroup[column]:
                    fig.add_trace(go.Scatter(x=data.index, y=data[column], name=column, yaxis=f"y{indexgroup+1}", legendgroup=legendgroup, hovertemplate="x: %{x:.2f}<br>y: %{y:.2f}", showlegend=False))
                    fig.layout[f"yaxis{indexgroup+1}"] = dict(title=legendgroup, anchor="x", overlaying="y", side="right", position=1, title_standoff=0, title_font= {"size" : sizep*rangemaxXaxis})
                else:
                    legendgroup = dicolistgroup[column]
                    indexgroup+=1
                    fig.add_trace(go.Scatter(x=[data.index[0]], y=[data[column][0]], name=legendgroup, marker=dict(color="Black", size=0.01), visible='legendonly', legendgroup=legendgroup, showlegend=True, yaxis=f"y{indexgroup+1}"))  #Q&D no better solution today
                    fig.add_trace(go.Scatter(x=data.index, y=data[column], name=column, yaxis=f"y{indexgroup+1}", legendgroup=legendgroup, hovertemplate="x: %{x:.2f}<br>y: %{y:.2f}", showlegend=False))
                    fig.layout[f"yaxis{indexgroup+1}"] = dict(title=legendgroup, anchor="free", overlaying="y", side="right", position=rangemaxXaxis + yspace*(indexgroup-1), title_standoff=0, title_font={"size" : sizep*rangemaxXaxis})
            else:
                if legendgroup == dicolistgroup[column]:
                    fig.add_trace(go.Scatter(x=data.index, y=data[column], name=column, yaxis=f"y{indexgroup+1}",legendgroup=legendgroup, hovertemplate="x: %{x:.2f}<br>y: %{y:.2f}", showlegend=False))
                    fig.layout[f"yaxis{indexgroup+1}"] = dict(title=legendgroup, anchor="free", overlaying="y", side="right", position=rangemaxXaxis + yspace*(indexgroup-1), title_standoff=0, title_font={"size" : sizep*rangemaxXaxis})
                else:
                    legendgroup = dicolistgroup[column]
                    indexgroup+=1
                    fig.add_trace(go.Scatter(x=[data.index[0]], y=[data[column][0]], name=legendgroup, marker=dict(color="Black", size=0.01), visible='legendonly', legendgroup=legendgroup, showlegend=True, yaxis=f"y{indexgroup+1}"))  #Q&D no better solution today
                    fig.add_trace(go.Scatter(x=data.index, y=data[column], name=column, yaxis=f"y{indexgroup+1}",legendgroup=legendgroup, hovertemplate="x: %{x:.2f}<br>y: %{y:.2f}", showlegend=False))
                    fig.layout[f"yaxis{indexgroup+1}"] = dict(title=legendgroup, anchor="free", overlaying="y", side="right", position=rangemaxXaxis + yspace*(indexgroup-1), title_standoff=0, title_font={"size" : sizep*rangemaxXaxis})
    else:
        rangemaxXaxis = defxaxislen(data.columns, yspace) #variable size of x axis
        assert rangemaxXaxis>0, "too many columns please drop useless columns and/or reduce yspace paramter and/or used groupby option"
        fig.update_layout(xaxis=dict(domain = [0,rangemaxXaxis], #place en % des x
                        title=data.index.name,
                        title_standoff=0,
                        ),
                        title_text="Fermentation multiple y-axes",
                        template="simple_white",
                        legend=dict(x=0,
                                    y=-0.1,
                                    tracegroupgap = 0,
                                    orientation="h",
                                    font=dict(
                                        family='Times New Roman',
                                        size = sizep,
                                        color='#000')
                                    ),
                        font=dict(
                            family='Linux Libertine, Times New Roman',
                            size = sizep,
                            color='#000'
                        ),
                        modebar_add=["v1hovermode", "toggleSpikeLines"],
                        )
        for index, column in enumerate(data.columns):
            if index==0:
                fig.add_trace(go.Scatter(x=data.index, y=data[column], name=column, hovertemplate="x: %{x:.2f}<br>y: %{y:.2f}"))
                fig.layout["yaxis"] = dict(title=column, title_standoff=0, title_font={"size" : sizep*rangemaxXaxis})
            elif index==1:
                fig.add_trace(go.Scatter(x=data.index, y=data[column], name=column, yaxis=f"y{index+1}", hovertemplate="x: %{x:.2f}<br>y: %{y:.2f}"))
                fig.layout[f"yaxis{index+1}"] = dict(title=column, anchor="x", overlaying="y", side="right", position=1, title_standoff=0, title_font= {"size" : sizep*rangemaxXaxis})
            else:
                fig.add_trace(go.Scatter(x=data.index, y=data[column], name=column, yaxis=f"y{index+1}", hovertemplate="x: %{x:.2f}<br>y: %{y:.2f}"))
                fig.layout[f"yaxis{index+1}"] = dict(title=column, anchor="free", overlaying="y", side="right", position=rangemaxXaxis + yspace*(index-1), title_standoff=0, title_font={"size" : sizep*rangemaxXaxis})
    if xtitle:
        fig.update_xaxes(title=xtitle)
    if path != "":
        if os.path.isfile(path):
            fig.write_html(os.path.splitext(path)[0]+"_multiyfgraph.html")
        elif os.path.isdir(path):
            fig.write_html(os.path.join(path,"fermy_multiyfgraph.html"))
        else:
            print("issue with path")
    else:
        fig.show()
    
def ccplot(data: pd.DataFrame, variableofinterest: str, path:str="", dicorename:Dict={}, forcepower:bool=False, addccviol:bool=False, colorby:Union[None,str]=None, outref:Union[None,pd.Series]=None) -> pd.DataFrame:
    """
    Function to compute and plot a simple (and pretty) Control Chart with Plotly express
    
    Input:
    -----
    Data will be columns values of the column named variableofinterest
    and index of the DataFrame as plot index
    
    Parameters:
    ----------
    path:str, default ""
    if provided the graph will be saved in this path
    
    dicorename: Dict, default {}
    allow users to custom names of axis with:
    {"old name" : "desired name", "old name2" : "desired name2"}
    
    forcepower: bool, default False
    allow users to force scientific notation 10^n for y axis
    and controls limits
    
    addccviol: bool, default False
    allow users to add a new column to original data with CC rules check-in 
    
    colorby: None or string
    allow users to color regarding column value
    
    outref: None or pd.Series
    use external data to compute ULC, LCL and mean
    
    Output:
    ------
    return a DataFrame like data with only Control Chart violations data
    """
    if type(outref) == pd.Series:
        center = outref.mean()
        std = outref.std()
    elif type(outref) == pd.DataFrame:
        print("outref have to be a pd.Series here it is a Dataframe")
        if outref.shape[1] == 1:
            print('outref was converted to pd.Series')
            outrefinserie = outref[outref.columns[0]]
            center = outrefinserie.mean()
            std = outrefinserie.std()
        else:
            print("outref can be converted to pf.Series only if shape = (n,1)")
            return None
    else:
        #basic data to compute LCL, center and UCL on data
        center = data[variableofinterest].mean()
        std = data[variableofinterest].std()
    nbsigma = [1,2,3]
    #color values add column in dataframe if value not ok regarding CC rules
    conditions = [  (data[variableofinterest]>=center+std*3),  # Out-of-control rule 1 > +3 sigma
                    (data[variableofinterest]<=center-std*3), # Out-of-control rule 1 < -3 sigma
                    ((data[variableofinterest] < center) & # Out-of-control rule 2 Nine points in a row -3 sigma
                    (data[variableofinterest].shift(1) < center) &
                    (data[variableofinterest].shift(2) < center) &
                    (data[variableofinterest].shift(3) < center) &
                    (data[variableofinterest].shift(4) < center) &
                    (data[variableofinterest].shift(5) < center) &
                    (data[variableofinterest].shift(6) < center) &
                    (data[variableofinterest].shift(7) < center) &
                    (data[variableofinterest].shift(8) < center)),
                    ((data[variableofinterest] > center) & # Out-of-control rule 2 Nine points in a row +3 sigma
                    (data[variableofinterest].shift(1) > center) &
                    (data[variableofinterest].shift(2) > center) &
                    (data[variableofinterest].shift(3) > center) &
                    (data[variableofinterest].shift(4) > center) &
                    (data[variableofinterest].shift(5) > center) &
                    (data[variableofinterest].shift(6) > center) &
                    (data[variableofinterest].shift(7) > center) &
                    (data[variableofinterest].shift(8) > center)),
                    ((data[variableofinterest].shift(2) >=center+std*2) & # Out-of-control rule 5 Two out of three points in a row +2 sigma
                    (data[variableofinterest].shift(1) >=center+std*2) &
                    (data[variableofinterest] >=center)),
                    ((data[variableofinterest].shift(2) >=center+std*2) & # Out-of-control rule 5 Two out of three points in a row +2 sigma
                    (data[variableofinterest] >=center+std*2) &
                    (data[variableofinterest].shift(1) >=center)),
                    ((data[variableofinterest].shift(2) <=center-std*2) & # Out-of-control rule 5 Two out of three points in a row -2 sigma
                    (data[variableofinterest].shift(1) <=center-std*2) &
                    (data[variableofinterest] <=center)),
                    ((data[variableofinterest].shift(2) <=center-std*2) & # Out-of-control rule 5 Two out of three points in a row -2 sigma
                    (data[variableofinterest] <=center-std*2) &
                    (data[variableofinterest].shift(1) <=center)),
                    ] # create a list of our conditions
    print("\nOut-of-control rule 1: > +3 sigma or < -3 sigma\nOut-of-control rule 2: Nine points in a row -3 sigma or +3 simga\nOut-of-control rule 5: Two out of three points in a row +2 sigma or -2 sigma\n")
    values = ['Out-of-control rule 1', 'Out-of-control rule 1', 'Out-of-control rule 2', 'Out-of-control rule 2', 'Out-of-control rule 5', 'Out-of-control rule 5', 'Out-of-control rule 5', 'Out-of-control rule 5'] # create a list of the values we want to assign for each condition
    #creat a new column with control status
    data['CClimitctrl'] = np.select(conditions, values, default = "Under control")
    # difine color map to unlight out-of-control
    color_discrete_map = {'Out-of-control rule 1': 'rgb(255,0,0)', 'Under control': 'rgb(31,119,180)', 'Out-of-control rule 2': 'rgb(217,75,83)' , 'Out-of-control rule 5' : 'rgb(238,25,154)'}
    #plot Control chart
    if not colorby:
        fig = px.scatter(data, x=data.index, y=variableofinterest,
                            #size  = variableofinterest, # size as function of a column
                            color = data.CClimitctrl, # color as function of a column
                            color_discrete_map=color_discrete_map, # set color according to user define map
                            #symbol = variableofinterest, # symbol as function of a column
                            title=f"Control Chart",
                            range_y = [center-6*std,center+6*std], # set default limit for Y
                            labels = dicorename,
                            template="simple_white"  # set easy template
                        )
    else:
        fig = px.scatter(data, x=data.index, y=variableofinterest,
                            #size  = variableofinterest, # size as function of a column
                            color = data[colorby], # color as function of a column
                            #symbol = variableofinterest, # symbol as function of a column
                            title=f"Control Chart",
                            range_y = [center-6*std,center+6*std], # set default limit for Y
                            labels = dicorename,
                            template="simple_white"  # set easy template
                        )
    #Add LCL UCL pretty print
    fig.add_hline(y=center, line_dash="longdashdot", line_color="blue", line_width=1, opacity=1)  #set center
    for sigma in nbsigma:
        if sigma<3:
            fig.add_hline(y=center+sigma*std, line_dash="dash", line_color="red", line_width=1, opacity = sigma/3)  # UCL
            fig.add_hline(y=center-sigma*std, line_dash="dash", line_color="red", line_width=1, opacity = sigma/3)  # LCL
        else:
            fig.add_hline(y=center+sigma*std, line_dash="longdash", line_color="red", line_width=1, opacity = sigma/3)  # UCL
            fig.add_hline(y=center-sigma*std, line_dash="longdash", line_color="red", line_width=1, opacity = sigma/3)  # LCL

    if len(data['CClimitctrl'].unique())<=1:
        fig.update(layout_showlegend=False) #hide legend if all point under control
        batchwithissues = pd.DataFrame()
    else:
        batchwithissues = data.loc[data['CClimitctrl']!="Under control",:]
        fig.update_layout(legend =dict(title=dict(text="Control Chart Violations")))
        
    if forcepower:
        fig.update_yaxes(exponentformat="power")  # force 10^n notation
                # add anotation ULC / AVG / LCL
        #ULC
        fig.add_annotation(
                x=1,
                y=center+3*std,
                xref="paper",
                yref="y",
                text=f"UCL={round(center+3*std,1):0.2E}",
                showarrow = False,
                yshift=10
                )
        #LCL
        fig.add_annotation(
                x=1,
                y=center-3*std,
                xref="paper",
                yref="y",
                text=f"LCL={round(center-3*std,1):0.2E}",
                showarrow = False,
                yshift=10
                )
        #center
        fig.add_annotation(
                x=1,
                y=center,
                xref="paper",
                yref="y",
                text=f"Avg={round(center,1):0.2E}",
                showarrow = False,
                yshift=10
                )
    else:
        # add anotation ULC / AVG / LCL
        #ULC
        fig.add_annotation(
                x=1,
                y=center+3*std,
                xref="paper",
                yref="y",
                text=f"UCL={round(center+3*std,1)}",
                showarrow = False,
                yshift=10
                )
        #LCL
        fig.add_annotation(
                x=1,
                y=center-3*std,
                xref="paper",
                yref="y",
                text=f"LCL={round(center-3*std,1)}",
                showarrow = False,
                yshift=10
                )
        #center
        fig.add_annotation(
                x=1,
                y=center,
                xref="paper",
                yref="y",
                text=f"Avg={round(center,1)}",
                showarrow = False,
                yshift=10
                )
                
    if path != "":
        fig.write_html(os.path.splitext(path)[0]+"_CCgraph.html")
    else:
        fig.show()
    if not addccviol:
        data.drop(columns="CClimitctrl", inplace=True)  # clean data add to graph violation
    return batchwithissues

pd.core.base.PandasObject.fplotsimple = basicplot
pd.core.base.PandasObject.fccplot = ccplot
pd.core.base.PandasObject.fplotmultiy = multiyplot

if __name__ == '__main__':
    """here argparser code"""
    parser = argparse.ArgumentParser(description = "Plotlyfermy add to Fermy method to plot fermentation data", 
    epilog="For Help or more information please contact Nicolas Hardy")
    
    parser.add_argument("filepath", metavar = "Root of data (Excel file)", type = str, help = "File with fermentation data first column have to be time (in datetime or float of hours)")
    parser.add_argument("-s","--save", action='store_true', dest="saved", help = "Option to save graph in the data folder")
    parser.add_argument("-cc", dest="variableofinterest", type = str, default = None, help = "Varaible of interest to use for the control chart creation")
    parser.add_argument('-d', '--dico', dest="dicorename", default= {}, nargs='*', action=ParseKwargs, help = "oldname1=newname1 oldname2=newname2 : is the renaming option format")
    
    
    args = parser.parse_args()
    filepath = args.filepath # filepath = "C:\\Users\\yp6247\\Desktop\\sampledata.xlsx"
    saved = args.saved # saved = True
    variableofinterest = args.variableofinterest
    dicorename = args.dicorename  # load dicorename
    """"code here"""
    data = pd.read_excel(filepath, index_col=0)  #load data with Pandas index have to be datetime.datetime
    if variableofinterest:
        if saved:
            data.ccplot(variableofinterest, path=filepath, dicorename=dicorename)
        else:
            data.ccplot(variableofinterest, dicorename=dicorename)
    else:
        if saved:
            data.mulyfplot(filepath)
        else:
            data.mulyfplot()
